"""Schema for the comparison object returned by the comparison command"""
from pydantic import BaseModel, Field, constr
from typing import List
from enum import Enum


class GeometryObjectTypes(str, Enum):
    """Types of Honeybee geometry objects."""
    shade = 'Shade'
    aperture = 'Aperture'
    door = 'Door'
    face = 'Face'
    room = 'Room'


class _DiffObjectBase(BaseModel):

    element_type: GeometryObjectTypes = Field(
        ...,
        description='Text for the type of object that has been changed.'
    )

    element_id: str = Field(
        ...,
        regex=r'^[^,;!\n\t]+$',
        min_length=1,
        max_length=100,
        description='Text string for the unique object ID that has changed.'
    )

    element_name: str = Field(
        None,
        description='Text string for the display name of the object that has changed.'
    )


class ChangedObject(_DiffObjectBase):

    type: constr(regex='^ChangedObject$') = 'ChangedObject'

    geometry_changed: bool = Field(
        ...,
        description='A boolean to note whether the geometry of the object has changed '
        '(True) or not (False). For the case of a Room, any change in the geometry of '
        'child Faces, Apertures or Doors will cause this property to be True. Note that '
        'this property is only True if the change in geometry produces a visible '
        'change greater than the base model tolerance. So converting the model '
        'between different unit systems, removing colinear vertices, or doing '
        'other transformations that are common for export to simulation engines '
        'will not trigger this property to become True.'
    )

    energy_changed: bool = Field(
        False,
        description='A boolean to note whether the energy properties of the object '
        'have changed (True) or not (False) such that it is possible for the '
        'properties of the changed object to be applied to the base model. '
        'For Rooms, this property will only be true if the energy property '
        'assigned to the Room has changed and will not be true if a property '
        'assigned to an individual child Face or Aperture has changed.'
    )

    radiance_changed: bool = Field(
        False,
        description='A boolean to note whether the radiance properties of the object '
        'have changed (True) or not (False) such that it is possible for the '
        'properties of the changed object to be applied to the base model. '
        'For Rooms, this property will only be true if the radiance property '
        'assigned to the Room has changed and will not be true if a property '
        'assigned to an individual child Face or Aperture has changed.'
    )

    geometry: List[dict] = Field(
        ...,
        description='A list of DisplayFace3D dictionaries for the new, changed '
        'geometry. The schema of DisplayFace3D can be found in the ladybug-display-'
        'schema documentation (https://www.ladybug.tools/ladybug-display-schema) '
        'and these objects can be used to generate visualizations of individual '
        'objects that have been changed. Note that this attribute is always '
        'included in the ChangedObject, even when geometry_changed is False.'
    )

    existing_geometry: List[dict] = Field(
        default=None,
        description='A list of DisplayFace3D dictionaries for the existing (base) '
        'geometry. The schema of DisplayFace3D can be found in the ladybug-display-'
        'schema documentation (https://www.ladybug.tools/ladybug-display-schema) '
        'and these objects can be used to generate visualizations of individual '
        'objects that have been changed. This attribute is optional and will '
        'NOT be output if geometry_changed is False.'
    )


class DeletedObject(_DiffObjectBase):

    type: constr(regex='^DeletedObject$') = 'DeletedObject'

    geometry: List[dict] = Field(
        ...,
        description='A list of DisplayFace3D dictionaries for the deleted '
        'geometry. The schema of DisplayFace3D can be found in the ladybug-display-'
        'schema documentation (https://www.ladybug.tools/ladybug-display-schema) '
        'and these objects can be used to generate visualizations of individual '
        'objects that have been deleted.'
    )


class AddedObject(_DiffObjectBase):

    type: constr(regex='^AddedObject$') = 'AddedObject'

    geometry: List[dict] = Field(
        ...,
        description='A list of DisplayFace3D dictionaries for the added '
        'geometry. The schema of DisplayFace3D can be found in the ladybug-display-'
        'schema documentation (https://www.ladybug.tools/ladybug-display-schema) '
        'and these objects can be used to generate visualizations of individual '
        'objects that have been added.'
    )


class ComparisonReport(BaseModel):

    type: constr(regex='^ComparisonReport$') = 'ComparisonReport'

    changed_objects: List[ChangedObject] = Field(
        default=None,
        description='A list of ChangedObject definitions for each top-level object '
        'that has changed in the model. To be a changed object, the object identifier '
        'must be the same in both models but some other property (either geometry '
        'or extension attributes) has experienced a meaningful change.'
    )

    deleted_objects: List[DeletedObject] = Field(
        default=None,
        description='A list of DeletedObject definitions for each top-level object '
        'that has been deleted in the process of going from the base model to the '
        'new model.'
    )

    added_objects: List[AddedObject] = Field(
        default=None,
        description='A list of AddedObject definitions for each top-level object '
        'that has been added in the process of going from the base model to the '
        'new model.'
    )


class ChangedInstruction(_DiffObjectBase):

    type: constr(regex='^ChangedInstruction$') = 'ChangedInstruction'

    update_geometry: bool = Field(
        True,
        description='A boolean to note whether the geometry of the object in the '
        'new/updated model should replace the base/existing geometry (True) or '
        'the existing geometry should be kept (False).'
    )

    update_energy: bool = Field(
        True,
        description='A boolean to note whether the energy properties of the '
        'object in the new/updated model should replace the base/existing energy '
        'properties (True) or the base/existing energy properties should '
        'be kept (False).'
    )

    update_radiance: bool = Field(
        True,
        description='A boolean to note whether the radiance properties of the '
        'object in the new/updated model should replace the base/existing radiance '
        'properties (True) or the base/existing radiance properties should '
        'be kept (False).'
    )


class DeletedInstruction(_DiffObjectBase):

    type: constr(regex='^DeletedInstruction$') = 'DeletedInstruction'


class AddedInstruction(_DiffObjectBase):

    type: constr(regex='^AddedInstruction$') = 'AddedInstruction'


class SyncInstructions(BaseModel):

    type: constr(regex='^SyncInstructions$') = 'SyncInstructions'

    changed_objects: List[ChangedInstruction] = Field(
        default=None,
        description='A list of ChangedInstruction definitions for each top-level '
        'object with properties to transfer from the new/updated model to the '
        'base/existing model.'
    )

    deleted_objects: List[DeletedInstruction] = Field(
        default=None,
        description='A list of DeletedInstruction definitions for each top-level object '
        'to be deleted from the base/existing model.'
    )

    added_objects: List[AddedInstruction] = Field(
        default=None,
        description='A list of AddedInstruction definitions for each top-level object '
        'to be added to the base/existing model from the new/updated model.'
    )



"""Schema for project information."""
from pydantic import BaseModel, Field, constr, AnyUrl
from typing import List, Union

from .altnumber import Autocalculate
from .energy.simulation import EfficiencyStandards, ClimateZones, BuildingTypes


class Location(BaseModel):
    """A Ladybug Location."""

    type: constr(regex='^Location$') = 'Location'

    city: str = Field(
        '-',
        description='Name of the city as a string.'
    )

    latitude: float = Field(
        0,
        description='Location latitude between -90 and 90 (Default: 0).'
    )

    longitude: float = Field(
        0,
        description='Location longitude between -180 (west) and 180 (east) (Default: 0).'
    )

    time_zone: Union[Autocalculate, int] = Field(
        Autocalculate(),
        ge=-12,
        le=14,
        description='Time zone between -12 hours (west) and +14 hours (east). '
        'If None, the time zone will be an estimated integer value derived from '
        'the longitude in accordance with solar time.'
    )

    elevation: float = Field(
        0,
        description='A number for elevation of the location in meters. (Default: 0).'
    )

    station_id: str = Field(
        None,
        description='ID of the location if the location is representing a weather '
        'station.'
    )

    source: str = Field(
        None, description='Source of data (e.g. TMY, TMY3).'
    )


class ProjectInfo(BaseModel):
    """Project information."""

    type: constr(regex='^ProjectInfo$') = 'ProjectInfo'

    north: float = Field(
        0,
        description='A number between -360 to 360 where positive values rotate the '
        'compass counterclockwise (towards the West) and negative values rotate the '
        'compass clockwise (towards the East).',
        le=360, ge=-360
    )

    weather_urls: List[AnyUrl] = Field(
        None,
        description='A list of URLs to zip files that includes EPW, DDY and STAT files. '
        'You can find these URLs from the EPWMAP. The first URL will be used as the '
        'primary weather file.'
    )

    location: Location = Field(
        None,
        description='Project location. This value is usually generated from the '
        'information in the weather files.'
    )

    ashrae_climate_zone: ClimateZones = Field(
        None,
        description='Project location climate zone.'
    )

    building_type: List[BuildingTypes] = Field(
        None,
        description='A list of building types for the project. The first building '
        'type is considered the primary type for the project.'
    )

    vintage: List[EfficiencyStandards] = Field(
        None,
        description='A list of building vintages (e.g. ASHRAE_2019, ASHRAE_2016).'
    )



"""Programtype Schema"""
from pydantic import Field, constr

from ._base import IDdEnergyBaseModel
from .load import PeopleAbridged, LightingAbridged, ElectricEquipmentAbridged, \
    GasEquipmentAbridged, ServiceHotWaterAbridged, \
    InfiltrationAbridged, VentilationAbridged, SetpointAbridged, \
    People, Lighting, ElectricEquipment, GasEquipment, ServiceHotWater, \
    Infiltration, Ventilation, Setpoint


class ProgramTypeAbridged(IDdEnergyBaseModel):

    type: constr(regex='^ProgramTypeAbridged$') = 'ProgramTypeAbridged'

    people: PeopleAbridged = Field(
        default=None,
        description='People to describe the occupancy of the program. If None, '
        'no occupancy will be assumed for the program.'
    )

    lighting: LightingAbridged = Field(
        default=None,
        description='Lighting to describe the lighting usage of the program. '
        'If None, no lighting will be assumed for the program.'
    )

    electric_equipment: ElectricEquipmentAbridged = Field(
        default=None,
        description='ElectricEquipment to describe the usage of electric equipment '
        'within the program. If None, no electric equipment will be assumed.'
    )

    gas_equipment: GasEquipmentAbridged = Field(
        default=None,
        description='GasEquipment to describe the usage of gas equipment '
        'within the program. If None, no gas equipment will be assumed.'
    )

    service_hot_water: ServiceHotWaterAbridged = Field(
        default=None,
        description='ServiceHotWater object to describe the usage of hot water '
        'within the program. If None, no hot water will be assumed.'
    )

    infiltration: InfiltrationAbridged = Field(
        default=None,
        description='Infiltration to describe the outdoor air leakage of '
        'the program. If None, no infiltration will be assumed for the program.'
    )

    ventilation: VentilationAbridged = Field(
        default=None,
        description='Ventilation to describe the minimum outdoor air requirement '
        'of the program. If None, no ventilation requirement will be assumed.'
    )

    setpoint: SetpointAbridged = Field(
        default=None,
        description='Setpoint object to describe the temperature and humidity setpoints '
        'of the program.  If None, the ProgramType cannot be assigned to a Room '
        'that is conditioned.'
    )


class ProgramType(ProgramTypeAbridged):

    type: constr(regex='^ProgramType$') = 'ProgramType'

    people: People = Field(
        default=None,
        description='People to describe the occupancy of the program. If None, '
        'no occupancy will be assumed for the program.'
    )

    lighting: Lighting = Field(
        default=None,
        description='Lighting to describe the lighting usage of the program. '
        'If None, no lighting will be assumed for the program.'
    )

    electric_equipment: ElectricEquipment = Field(
        default=None,
        description='ElectricEquipment to describe the usage of electric equipment '
        'within the program. If None, no electric equipment will be assumed.'
    )

    gas_equipment: GasEquipment = Field(
        default=None,
        description='GasEquipment to describe the usage of gas equipment '
        'within the program. If None, no gas equipment will be assumed.'
    )

    service_hot_water: ServiceHotWater = Field(
        default=None,
        description='ServiceHotWater object to describe the usage of hot water '
        'within the program. If None, no hot water will be assumed.'
    )

    infiltration: Infiltration = Field(
        default=None,
        description='Infiltration to describe the outdoor air leakage of '
        'the program. If None, no infiltration will be assumed for the program.'
    )

    ventilation: Ventilation = Field(
        default=None,
        description='Ventilation to describe the minimum outdoor air requirement '
        'of the program. If None, no ventilation requirement will be assumed.'
    )

    setpoint: Setpoint = Field(
        default=None,
        description='Setpoint object to describe the temperature and humidity setpoints '
        'of the program.  If None, the ProgramType cannot be assigned to a Room '
        'that is conditioned.'
    )



"""Load Schemas"""
from pydantic import Field, root_validator, constr
from typing import Union
from enum import Enum

from ._base import IDdEnergyBaseModel
from .schedule import ScheduleRuleset, ScheduleFixedInterval
from ..altnumber import Autocalculate


class PeopleAbridged(IDdEnergyBaseModel):

    type: constr(regex='^PeopleAbridged$') = 'PeopleAbridged'

    people_per_area: float = Field(
        ...,
        ge=0,
        description='People per floor area expressed as [people/m2]'
    )

    occupancy_schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of a schedule for the occupancy over the course of the '
        'year. The type of this schedule should be Fractional and the fractional '
        'values will get multiplied by the people_per_area to yield a complete '
        'occupancy profile.'
    )

    activity_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a schedule for the activity of the occupants over '
        'the course of the year. The type of this schedule should be ActivityLevel '
        'and the values of the schedule equal to the number of Watts given off by an '
        'individual person in the room. If None, a default constant schedule with '
        '120 Watts per person will be used, which is typical of awake, adult humans '
        'who are seated.'
    )

    radiant_fraction: float = Field(
        0.3,
        ge=0,
        le=1,
        description='The radiant fraction of sensible heat released by people. '
        '(Default: 0.3).'
    )

    latent_fraction: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        le=1,
        description='Number for the latent fraction of heat gain due to people or '
        'an Autocalculate object.'
    )

    @root_validator
    def check_sum_fractions(cls, values):
        "Ensure sum is less than 1."
        rad = values.get('radiant_fraction')
        latent = values.get('latent_fraction')
        if latent is not None and isinstance(latent, float):
            assert rad + latent <= 1, \
                'Sum of radiant and latent fractions cannot be greater than 1.'
        return values


class People(PeopleAbridged):

    type: constr(regex='^People$') = 'People'

    occupancy_schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='A schedule for the occupancy over the course of the '
        'year. The type of this schedule should be Fractional and the fractional '
        'values will get multiplied by the people_per_area to yield a complete '
        'occupancy profile.'
    )

    activity_schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        default=None,
        description='A schedule for the activity of the occupants over the '
        'course of the year. The type of this schedule should be ActivityLevel '
        'and the values of the schedule equal to the number of Watts given off by an '
        'individual person in the room. If None, a default constant schedule with '
        '120 Watts per person will be used, which is typical of awake, adult humans '
        'who are seated.'
    )


class LightingAbridged(IDdEnergyBaseModel):

    type: constr(regex='^LightingAbridged$') = 'LightingAbridged'

    watts_per_area: float = Field(
        ...,
        ge=0,
        description='Lighting per floor area as [W/m2].'
    )

    schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the use of lights over the course '
        'of the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the watts_per_area to yield a '
        'complete lighting profile.'
    )

    visible_fraction: float = Field(
        0.25,
        ge=0,
        le=1,
        description='The fraction of heat from lights that goes into the zone as '
        'visible (short-wave) radiation. (Default: 0.25).'
    )

    radiant_fraction: float = Field(
        0.32,
        ge=0,
        le=1,
        description='The fraction of heat from lights that is long-wave radiation. '
        '(Default: 0.32).'
    )

    return_air_fraction: float = Field(
        0.0,
        ge=0,
        le=1,
        description='The fraction of the heat from lights that goes into the zone '
        'return air. (Default: 0).'
    )

    baseline_watts_per_area: float = Field(
        11.84029,
        ge=0,
        description='The baseline lighting power density in [W/m2] of floor area. '
        'This baseline is useful to track how much better the installed lights are '
        'in comparison to a standard like ASHRAE 90.1. If set to None, it will '
        'default to 11.84029 W/m2, which is that ASHRAE 90.1-2004 baseline for '
        'an office.'
    )

    @root_validator
    def check_sum_fractions(cls, values):
        "Ensure sum is less than 1."
        return_air = values.get('return_air_fraction')
        vis = values.get('visible_fraction')
        rad = values.get('radiant_fraction')
        assert sum((return_air, vis, rad)) <= 1, \
            'Sum of visible, radiant, and return air fractions cannot be greater than 1.'
        return values


class Lighting(LightingAbridged):

    type: constr(regex='^Lighting$') = 'Lighting'

    schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='The schedule for the use of lights over the course of '
        'the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the watts_per_area to yield a '
        'complete lighting profile.'
    )


class _EquipmentBase(IDdEnergyBaseModel):

    watts_per_area: float = Field(
        ...,
        ge=0,
        description='Equipment level per floor area as [W/m2].'
    )

    schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the use of equipment over the '
        'course of the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the watts_per_area to yield '
        'a complete equipment profile.'
    )

    radiant_fraction: float = Field(
        0,
        ge=0,
        le=1,
        description='Number for the amount of long-wave radiation heat given off'
        ' by equipment. Default value is 0.'
    )

    latent_fraction: float = Field(
        0,
        ge=0,
        le=1,
        description='Number for the amount of latent heat given off by '
        'equipment. Default value is 0.'

    )

    lost_fraction: float = Field(
        0,
        ge=0,
        le=1,
        description='Number for the amount of “lost” heat being given off by '
        'equipment. The default value is 0.'
    )

    @root_validator
    def check_sum_fractions(cls, values):
        "Ensure sum is less than 1."
        rad = values.get('radiant_fraction')
        latent = values.get('latent_fraction')
        lost = values.get('lost_fraction')
        assert sum((rad, latent, lost)) <= 1, \
            'Sum of radiant, latent, and lost fractions cannot be greater than 1.'
        return values


class ElectricEquipmentAbridged(_EquipmentBase):

    type: constr(regex='^ElectricEquipmentAbridged$') = 'ElectricEquipmentAbridged'


class ElectricEquipment(ElectricEquipmentAbridged):

    type: constr(regex='^ElectricEquipment$') = 'ElectricEquipment'

    schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='The schedule for the use of equipment over the course '
        'of the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the watts_per_area to yield '
        'a complete equipment profile.'
    )


class GasEquipmentAbridged(_EquipmentBase):

    type: constr(regex='^GasEquipmentAbridged$') = 'GasEquipmentAbridged'


class GasEquipment(GasEquipmentAbridged):

    type: constr(regex='^GasEquipment$') = 'GasEquipment'

    schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='The schedule for the use of equipment over the course '
        'of the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the watts_per_area to yield '
        'a complete equipment profile.'
    )


class ServiceHotWaterAbridged(IDdEnergyBaseModel):

    type: constr(regex='^ServiceHotWaterAbridged$') = 'ServiceHotWaterAbridged'

    flow_per_area: float = Field(
        ...,
        ge=0,
        description='Number for the total volume flow rate of water per unit area '
        'of floor [L/h-m2].'
    )

    schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the hot water use over the course '
        'of the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the flow_per_area to yield a '
        'complete water usage profile.'
    )

    target_temperature: float = Field(
        60,
        gt=0,
        description='Number for the target temperature of water out of the tap (C). '
        'This the temperature after hot water has been mixed with cold water '
        'from the water mains. The default is 60C, which essentially assumes that the '
        'flow_per_area on this object is only for water straight out of the '
        'water heater.'
    )

    sensible_fraction: float = Field(
        0.2,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the fraction of the total hot water '
        'load given off as sensible heat in the zone.'
    )

    latent_fraction: float = Field(
        0.05,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the fraction of the total hot '
        'water load that is latent.'
    )

    @root_validator
    def check_sum_fractions(cls, values):
        "Ensure sum is less than 1."
        sens = values.get('sensible_fraction')
        lat = values.get('latent_fraction')
        assert sum((sens, lat)) <= 1, \
            'Sum of sensible and latent fractions cannot be greater than 1.'
        return values


class ServiceHotWater(ServiceHotWaterAbridged):

    type: constr(regex='^ServiceHotWater$') = 'ServiceHotWater'

    schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='The schedule for the use of hot water over the course of '
        'the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the flow_per_area to yield a '
        'complete water usage profile.'
    )


class InfiltrationAbridged(IDdEnergyBaseModel):

    type: constr(regex='^InfiltrationAbridged$') = 'InfiltrationAbridged'

    flow_per_exterior_area: float = Field(
        ...,
        ge=0,
        description='Number for the infiltration per exterior surface area in m3/s-m2.'
    )

    schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the infiltration over the course of '
        'the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the flow_per_exterior_area '
        'to yield a complete infiltration profile.'
    )

    constant_coefficient: float = Field(
        1,
        ge=0
    )

    temperature_coefficient: float = Field(
        0,
        ge=0
    )

    velocity_coefficient: float = Field(
        0,
        ge=0
    )


class Infiltration(InfiltrationAbridged):

    type: constr(regex='^Infiltration$') = 'Infiltration'

    schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='The schedule for the infiltration over the course of '
        'the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the flow_per_exterior_area '
        'to yield a complete infiltration profile.'
    )


class VentilationAbridged(IDdEnergyBaseModel):

    type: constr(regex='^VentilationAbridged$') = 'VentilationAbridged'

    flow_per_person: float = Field(
        0,
        ge=0,
        description='Intensity of ventilation in[] m3/s per person]. Note that '
        'setting this value does not mean that ventilation is varied based on '
        'real-time occupancy but rather that the design level of ventilation '
        'is determined using this value and the People object of the Room.'
    )

    flow_per_area: float = Field(
        0,
        ge=0,
        description='Intensity of ventilation in [m3/s per m2 of floor area].'
    )

    air_changes_per_hour: float = Field(
        0,
        ge=0,
        description='Intensity of ventilation in air changes per hour (ACH) for '
        'the entire Room.'
    )

    flow_per_zone: float = Field(
        0,
        ge=0,
        description='Intensity of ventilation in m3/s for the entire Room.'
    )

    schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the ventilation over the course of '
        'the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the total design flow rate '
        '(determined from the sum of the other 4 fields) to yield a complete '
        'ventilation profile.'
    )


class Ventilation(VentilationAbridged):

    type: constr(regex='^Ventilation$') = 'Ventilation'

    schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        default=None,
        description='Schedule for the ventilation over the course of '
        'the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the total design flow rate '
        '(determined from the sum of the other 4 fields) to yield a complete '
        'ventilation profile.'
    )


class SetpointAbridged(IDdEnergyBaseModel):
    """Used to specify information about the setpoint schedule."""

    type: constr(regex='^SetpointAbridged$') = 'SetpointAbridged'

    cooling_schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the cooling setpoint. The values in '
        'this schedule should be temperature in [C].'
    )

    heating_schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the heating setpoint. The values in '
        'this schedule should be temperature in [C].'
    )

    humidifying_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the humidification setpoint. '
        'The values in this schedule should be in [%].'
    )

    dehumidifying_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the dehumidification setpoint. '
        'The values in this schedule should be in [%].'
    )

    setpoint_cutout_difference: float = Field(
        0,
        ge=0,
        description='An optional positive number for the temperature '
        'difference between the cutout temperature and the setpoint temperature. '
        'Specifying a non-zero number here is useful for modeling the throttling '
        'range associated with a given setup of setpoint controls and HVAC equipment. '
        'Throttling ranges describe the range where a zone is slightly over-cooled '
        'or over-heated beyond the thermostat setpoint. They are used to avoid '
        'situations where HVAC systems turn on only to turn off a few minutes later, '
        'thereby wearing out the parts of mechanical systems faster. They can '
        'have a minor impact on energy consumption and can often have significant '
        'impacts on occupant thermal comfort, though using the default value '
        'of zero will often yield results that are close enough when trying '
        'to estimate the annual heating/cooling energy use. Specifying a value '
        'of zero effectively assumes that the system will turn on whenever '
        'conditions are outside the setpoint range and will cut out as soon '
        'as the setpoint is reached.'
    )

    @root_validator
    def check_both_humid_sch(cls, values):
        "Ensure that the other humidity schedule is included when one is."
        humid = values.get('humidifying_schedule')
        dehumid = values.get('dehumidifying_schedule')
        if humid is not None:
            assert dehumid is not None, 'When humidifying_schedule is specified, ' \
                'dehumidifying_schedule must also be specified.'
        if dehumid is not None:
            assert humid is not None, 'When dehumidifying_schedule is specified, ' \
                'humidifying_schedule must also be specified.'
        return values


class Setpoint(SetpointAbridged):
    """Used to specify information about the setpoint schedule."""

    type: constr(regex='^Setpoint$') = 'Setpoint'

    cooling_schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='Schedule for the cooling setpoint. The values in '
        'this schedule should be temperature in [C].'
    )

    heating_schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='Schedule for the heating setpoint. The values in '
        'this schedule should be temperature in [C].'
    )

    humidifying_schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        default=None,
        description='Schedule for the humidification setpoint. The values '
        'in this schedule should be in [%].'
    )

    dehumidifying_schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        default=None,
        description='Schedule for the dehumidification setpoint. The values '
        'in this schedule should be in [%].'
    )


class FuelTypes (str, Enum):
    """Designates the acceptable fuel types for process loads."""
    electricity = 'Electricity'
    natural_gas = 'NaturalGas'
    propane = 'Propane'
    fuel_oil_no_1 = 'FuelOilNo1'
    fuel_oil_no_2 = 'FuelOilNo2'
    diesel = 'Diesel'
    gasoline = 'Gasoline'
    coal = 'Coal'
    steam = 'Steam'
    district_heating = 'DistrictHeating'
    district_cooling = 'DistrictCooling'
    other_fuel_1 = 'OtherFuel1'
    other_fuel_2 = 'OtherFuel2'
    none = 'None'


class ProcessAbridged(IDdEnergyBaseModel):

    type: constr(regex='^ProcessAbridged$') = 'ProcessAbridged'

    watts: float = Field(
        ...,
        ge=0,
        description='A number for the process load power in Watts.'
    )

    schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the use of the process over the '
        'course of the year. The type of this schedule should be Fractional and the '
        'fractional values will get multiplied by the watts to yield a complete '
        'equipment profile.'
    )

    fuel_type: FuelTypes = Field(
        ...,
        description='Text to denote the type of fuel consumed by the process. '
        'Using the "None" type indicates that no end uses will be associated '
        'with the process, only the zone gains.'
    )

    end_use_category: str = Field(
        'Process',
        min_length=1,
        max_length=100,
        description='Text to indicate the end-use subcategory, which will identify '
        'the process load in the end use output. For example, “Cooking”, '
        '“Clothes Drying”, etc. A new meter for reporting is created for each '
        'unique subcategory.'
    )

    radiant_fraction: float = Field(
        0,
        ge=0,
        le=1,
        description='Number for the amount of long-wave radiation heat given off'
        ' by the process load. Default value is 0.'
    )

    latent_fraction: float = Field(
        0,
        ge=0,
        le=1,
        description='Number for the amount of latent heat given off by the process '
        'load. Default value is 0.'

    )

    lost_fraction: float = Field(
        0,
        ge=0,
        le=1,
        description='Number for the amount of “lost” heat being given off by '
        'the process load. The default value is 0.'
    )

    @root_validator
    def check_sum_fractions(cls, values):
        "Ensure sum is less than 1."
        rad = values.get('radiant_fraction')
        latent = values.get('latent_fraction')
        lost = values.get('lost_fraction')
        assert sum((rad, latent, lost)) <= 1, \
            'Sum of radiant, latent, and lost fractions cannot be greater than 1.'
        return values



"""Ventilative cooling Schema"""
from pydantic import Field, constr

from .._base import NoExtraBaseModel
from ._base import EnergyBaseModel
from enum import Enum


class VentilationControlAbridged(NoExtraBaseModel):

    type: constr(regex='^VentilationControlAbridged$') = 'VentilationControlAbridged'

    min_indoor_temperature: float = Field(
        -100,
        ge=-100,
        le=100,
        description='A number for the minimum indoor temperature at which to '
        'ventilate in Celsius. Typically, this variable is used to initiate ventilation.'
    )

    max_indoor_temperature: float = Field(
        100,
        ge=-100,
        le=100,
        description='A number for the maximum indoor temperature at which to '
        'ventilate in Celsius. This can be used to set a maximum temperature at '
        'which point ventilation is stopped and a cooling system is turned on.'
    )

    min_outdoor_temperature: float = Field(
        -100,
        ge=-100,
        le=100,
        description='A number for the minimum outdoor temperature at which to ventilate '
        'in Celsius. This can be used to ensure ventilative cooling does not happen '
        'during the winter even if the Room is above the min_indoor_temperature.'
    )

    max_outdoor_temperature: float = Field(
        100,
        ge=-100,
        le=100,
        description='A number for the maximum indoor temperature at which to ventilate '
        'in Celsius. This can be used to set a limit for when it is considered too hot '
        'outside for ventilative cooling.'
    )

    delta_temperature: float = Field(
        -100,
        ge=-100,
        le=100,
        description='A number for the temperature differential in Celsius between '
        'indoor and outdoor below which ventilation is shut off.  This should usually '
        'be a positive number so that ventilation only occurs when the outdoors is '
        'cooler than the indoors. Negative numbers indicate how much hotter the '
        'outdoors can be than the indoors before ventilation is stopped.'
    )

    schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of the schedule for the ventilation over the course of '
        'the year. Note that this is applied on top of any setpoints. The type of this '
        'schedule should be On/Off and values should be either 0 (no possibility of '
        'ventilation) or 1 (ventilation possible).'
    )


class VentilationType(str, Enum):
    exhaust = 'Exhaust'
    intake = 'Intake'
    balanced = 'Balanced'


class VentilationFan(EnergyBaseModel):

    type: constr(regex='^VentilationFan$') = 'VentilationFan'

    flow_rate: float = Field(
        ...,
        gt=0,
        description='A number for the flow rate of the fan in m3/s.'
    )

    ventilation_type: VentilationType = Field(
        default=VentilationType.balanced,
        description='Text to indicate the type of type of ventilation. Choose '
        'from the options below. For either Exhaust or Intake, values for '
        'fan pressure and efficiency define the fan electric consumption. For Exhaust '
        'ventilation, the conditions of the air entering the space are assumed '
        'to be equivalent to outside air conditions. For Intake and Balanced '
        'ventilation, an appropriate amount of fan heat is added to the '
        'entering air stream. For Balanced ventilation, both an intake fan '
        'and an exhaust fan are assumed to co-exist, both having the same '
        'flow rate and power consumption (using the entered values for fan '
        'pressure rise and fan total efficiency). Thus, the fan electric '
        'consumption for Balanced ventilation is twice that for the Exhaust or '
        'Intake ventilation types which employ only a single fan.'
    )

    pressure_rise: float = Field(
        ...,
        gt=0,
        description='A number for the the pressure rise across the fan in Pascals '
        '(N/m2). This is often a function of the fan speed and the conditions in '
        'which the fan is operating since having the fan blow air through filters '
        'or narrow ducts will increase the pressure rise that is needed to '
        'deliver the input flow rate. The pressure rise plays an important role in '
        'determining the amount of energy consumed by the fan. Smaller fans like '
        'a 0.05 m3/s desk fan tend to have lower pressure rises around 60 Pa. '
        'Larger fans, such as a 6 m3/s fan used for ventilating a large room tend '
        'to have higher pressure rises around 400 Pa. The highest pressure rises '
        'are typically for large fans blowing air through ducts and filters, which '
        'can have pressure rises as high as 1000 Pa.'
    )

    efficiency: float = Field(
        ...,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the overall efficiency of the fan. '
        'Specifically, this is the ratio of the power delivered to the fluid '
        'to the electrical input power. It is the product of the fan motor '
        'efficiency and the fan impeller efficiency. Fans that have a higher blade '
        'diameter and operate at lower speeds with smaller pressure rises for '
        'their size tend to have higher efficiencies. Because motor efficiencies '
        'are typically between 0.8 and 0.9, the best overall fan efficiencies '
        'tend to be around 0.7 with most typical fan efficiencies between 0.5 and '
        '0.7. The lowest efficiencies often happen for small fans in situations '
        'with high pressure rises for their size, which can result in efficiencies '
        'as low as 0.15.'
    )

    control: VentilationControlAbridged = Field(
        default=None,
        description='A VentilationControl object that dictates the conditions under '
        'which the fan is turned on. If None, a default VentilationControl will '
        'be generated, which will keep the fan on all of the time.'
    )


class VentilationOpening(NoExtraBaseModel):

    type: constr(regex='^VentilationOpening$') = 'VentilationOpening'

    fraction_area_operable: float = Field(
        0.5,
        ge=0,
        le=1,
        description='A number for the fraction of the window area that is operable.'
    )

    fraction_height_operable: float = Field(
        1.0,
        ge=0,
        le=1,
        description='A number for the fraction of the distance from the bottom of the '
        'window to the top that is operable'
    )

    discharge_coefficient: float = Field(
        0.45,
        ge=0,
        le=1,
        description='A number that will be multiplied by the area of the window in the '
        'stack (buoyancy-driven) part of the equation to account for additional '
        'friction from window geometry, insect screens, etc. Typical values include '
        '0.45, for unobstructed windows WITH insect screens and 0.65 for unobstructed '
        'windows WITHOUT insect screens. This value should be lowered if windows are '
        'of an awning or casement type and are not allowed to fully open.'
    )

    wind_cross_vent: bool = Field(
        False,
        description='Boolean to indicate if there is an opening of roughly equal area '
        'on the opposite side of the Room such that wind-driven cross ventilation will '
        'be induced. If False, the assumption is that the operable area is primarily on '
        'one side of the Room and there is no wind-driven ventilation.'
    )

    flow_coefficient_closed: float = Field(
        default=0,
        ge=0,
        description='An optional number in kg/s-m, at 1 Pa per meter of crack length, '
        'used to calculate the mass flow rate when the opening is closed; required to '
        'run an AirflowNetwork simulation. The DesignBuilder Cracks template defines '
        'the flow coefficient for a tight, low-leakage closed external window to be '
        '0.00001, and the flow coefficient for a very poor, high-leakage closed '
        'external window to be 0.003.'
    )

    flow_exponent_closed: float = Field(
        default=0.65,
        ge=0.5,
        le=1,
        description='An optional dimensionless number between 0.5 and 1 used to '
        'calculate the mass flow rate when the opening is closed; required to run an '
        'AirflowNetwork simulation. This value represents the leak geometry impact '
        'on airflow, with 0.5 generally corresponding to turbulent orifice flow and 1 '
        'generally corresponding to laminar flow. The default of 0.65 is '
        'representative of many cases of wall and window leakage, used when the '
        'exponent cannot be measured.'
    )

    two_way_threshold: float = Field(
        default=0.0001,
        gt=0,
        description='A number in kg/m3 indicating the minimum density difference above '
        'which two-way flow may occur due to stack effect, required to run an '
        'AirflowNetwork simulation. This value is required because the air density '
        'difference between two zones (which drives two-way air flow) will tend '
        'towards division by zero errors as the air density difference approaches '
        'zero. The default of 0.0001 is a typical default value used for AirflowNetwork '
        'openings.'
    )


class AFNCrack(NoExtraBaseModel):
    """Properties for airflow through a crack."""

    type: constr(regex='^AFNCrack$') = 'AFNCrack'

    flow_coefficient: float = Field(
        ...,
        gt=0,
        description='A number in kg/s-m at 1 Pa per meter of crack length at the '
        'conditions defined in the ReferenceCrack condition; required to run an '
        'AirflowNetwork simulation. The DesignBuilder Cracks template defines the flow '
        'coefficient for a tight, low-leakage wall to be 0.00001 and 0.001 for external '
        'and internal constructions, respectively. Flow coefficients for a very poor, '
        'high-leakage wall are defined to be 0.0004 and 0.019 for external and internal '
        'constructions, respectively.'
    )

    flow_exponent: float = Field(
        default=0.65,
        ge=0.5,
        le=1,
        description='An optional dimensionless number between 0.5 and 1 used to '
        'calculate the crack mass flow rate; required to run an '
        'AirflowNetwork simulation. This value represents the leak geometry impact '
        'on airflow, with 0.5 generally corresponding to turbulent orifice flow and 1 '
        'generally corresponding to laminar flow. The default of 0.65 is '
        'representative of many cases of wall and window leakage, used when the '
        'exponent cannot be measured.'
    )


class VentilationControlType(str, Enum):
    single_zone = 'SingleZone'
    multi_zone_with_distribution = 'MultiZoneWithDistribution'
    multi_zone_without_distribution = 'MultiZoneWithoutDistribution'


class BuildingType(str, Enum):
    lowrise = 'LowRise'
    highrise = 'HighRise'


class VentilationSimulationControl(NoExtraBaseModel):
    """The global parameters used in the ventilation simulation."""

    type: constr(regex='^VentilationSimulationControl$') = 'VentilationSimulationControl'

    vent_control_type: VentilationControlType = Field(
        default=VentilationControlType.single_zone,
        description='Text indicating type of ventilation control. Choices are: '
        'SingleZone, MultiZoneWithDistribution, MultiZoneWithoutDistribution. The '
        'MultiZone options will model air flow with the AirflowNetwork model, which '
        'is generally more accurate then the SingleZone option, but will take '
        'considerably longer to simulate, and requires defining more ventilation '
        'parameters to explicitly account for weather and building-induced pressure '
        'differences, and the leakage geometry corresponding to specific windows, '
        'doors, and surface cracks.'
    )

    reference_temperature: float = Field(
        default=20,
        ge=-273.15,
        description='Reference temperature measurement in Celsius under which the '
        'surface crack data were obtained.'
    )

    reference_pressure: float = Field(
        default=101325,
        ge=31000,
        le=120000,
        description='Reference barometric pressure measurement in Pascals under which '
        'the surface crack data were obtained.'
    )

    reference_humidity_ratio: float = Field(
        default=0,
        ge=0,
        description='Reference humidity ratio measurement in kgWater/kgDryAir under '
        'which the surface crack data were obtained.'
    )

    building_type: BuildingType = Field(
        default=BuildingType.lowrise,
        description='Text indicating relationship between building footprint and '
        'height used to calculate the wind pressure coefficients for exterior surfaces.'
        'Choices are: LowRise and HighRise. LowRise corresponds to rectangular building '
        'whose height is less then three times the width and length of the footprint. '
        'HighRise corresponds to a rectangular building whose height is more than three '
        'times the width and length of the footprint. This parameter is required to '
        'automatically calculate wind pressure coefficients for the AirflowNetwork '
        'simulation. If used for complex building geometries that cannot be described '
        'as a highrise or lowrise rectangular mass, the resulting air flow and pressure '
        'simulated on the building surfaces may be inaccurate.'
    )

    long_axis_angle: float = Field(
        default=0,
        ge=0,
        le=180,
        description='The clockwise rotation in degrees from true North of the long axis '
        'of the building. This parameter is required to automatically calculate wind '
        'pressure coefficients for the AirflowNetwork simulation. If used for complex '
        'building geometries that cannot be described as a highrise or lowrise '
        'rectangular mass, the resulting air flow and pressure simulated on the '
        'building surfaces may be inaccurate.'
    )

    aspect_ratio: float = Field(
        default=1,
        gt=0,
        le=1,
        description='Aspect ratio of a rectangular footprint, defined as the ratio of '
        'length of the short axis divided by the length of the long axis. This '
        'parameter is required to automatically calculate wind '
        'pressure coefficients for the AirflowNetwork simulation. If used for complex '
        'building geometries that cannot be described as a highrise or lowrise '
        'rectangular mass, the resulting air flow and pressure simulated on the '
        'building surfaces may be inaccurate.'
    )



"""Changes associated with version Honeybee schema version 1.39.12."""


def version_1_39_12(model_dict):
    """Implement changes in a Model dict to make it compatible with version 1.39.12."""
    removed_equip = 'PSZ-AC district chilled water with baseboard district hot water'
    replaced_equip = 'PSZ-AC district chilled water with district hot water'
    if 'energy' in model_dict['properties']:
        if 'hvacs' in model_dict['properties']['energy']:
            for hvac in model_dict['properties']['energy']['hvacs']:
                if hvac['type'] != 'IdealAirSystemAbridged' and \
                        hvac['equipment_type'] == removed_equip:
                    hvac['equipment_type'] = replaced_equip
    return model_dict



"""Heating/cooling systems without any ventilation."""
from pydantic import Field, constr
from enum import Enum

from ._template import _TemplateSystem, RadiantFaceTypes


class _HeatCoolBase(_TemplateSystem):
    """Base class for all heating/cooling systems without any ventilation.

    These systems are only designed to satisfy heating + cooling demand and they
    cannot meet any minimum ventilation requirements.

    As such, these systems tend to be used in residential or storage settings where
    meeting minimum ventilation requirements may not be required or the density
    of occupancy is so low that infiltration is enough to meet fresh air demand.
    """


class FCUEquipmentType(str, Enum):
    fcu_chill_gb = 'FCU_Chiller_Boiler'
    fcu_chill_ashp = 'FCU_Chiller_ASHP'
    fcu_chill_dhw = 'FCU_Chiller_DHW'
    fcu_chill_base = 'FCU_Chiller_ElectricBaseboard'
    fcu_chill_guh = 'FCU_Chiller_GasHeaters'
    fcu_chill = 'FCU_Chiller'
    fcu_ac_chill_gb = 'FCU_ACChiller_Boiler'
    fcu_ac_chill_ashp = 'FCU_ACChiller_ASHP'
    fcu_ac_chill_dhw = 'FCU_ACChiller_DHW'
    fcu_ac_chill_base = 'FCU_ACChiller_ElectricBaseboard'
    fcu_ac_chill_guh = 'FCU_ACChiller_GasHeaters'
    fcu_ac_chill = 'FCU_ACChiller'
    fcu_dcw_gb = 'FCU_DCW_Boiler'
    fcu_dcw_ashp = 'FCU_DCW_ASHP'
    fcu_dcw_dhw = 'FCU_DCW_DHW'
    fcu_dcw_base = 'FCU_DCW_ElectricBaseboard'
    fcu_dcw_guh = 'FCU_DCW_GasHeaters'
    fcu_dcw = 'FCU_DCW'


class BaseboardEquipmentType(str, Enum):
    e_base = 'ElectricBaseboard'
    gb_base = 'BoilerBaseboard'
    ashp_base = 'ASHPBaseboard'
    dhw_base = 'DHWBaseboard'


class EvaporativeCoolerEquipmentType(str, Enum):
    evap_e_base = 'EvapCoolers_ElectricBaseboard'
    evap_gb_base = 'EvapCoolers_BoilerBaseboard'
    evap_ashp_base = 'EvapCoolers_ASHPBaseboard'
    evap_dhw_base = 'EvapCoolers_DHWBaseboard'
    evap_furnace = 'EvapCoolers_Furnace'
    evap_guh = 'EvapCoolers_UnitHeaters'
    evap = 'EvapCoolers'


class WSHPEquipmentType(str, Enum):
    wshp_fc_gb = 'WSHP_FluidCooler_Boiler'
    wshp_ct_gb = 'WSHP_CoolingTower_Boiler'
    wshp_gshp = 'WSHP_GSHP'
    wshp_dcw_dhw = 'WSHP_DCW_DHW'


class ResidentialEquipmentType(str, Enum):
    res_ac_e_base = 'ResidentialAC_ElectricBaseboard'
    res_ac_gb_base = 'ResidentialAC_BoilerBaseboard'
    res_ac_ashp_base = 'ResidentialAC_ASHPBaseboard'
    res_ac_dhw_base = 'ResidentialAC_DHWBaseboard'
    res_ac_furnace = 'ResidentialAC_ResidentialFurnace'
    res_ac = 'ResidentialAC'
    res_hp = 'ResidentialHP'
    res_hp_no_cool = 'ResidentialHPNoCool'
    res_furnace = 'ResidentialFurnace'


class WindowACEquipmentType(str, Enum):
    win_ac_e_base = 'WindowAC_ElectricBaseboard'
    win_ac_gb_base = 'WindowAC_BoilerBaseboard'
    win_ac_ashp_base = 'WindowAC_ASHPBaseboard'
    win_ac_dhw_base = 'WindowAC_DHWBaseboard'
    win_ac_furnace = 'WindowAC_Furnace'
    win_ac_guh = 'WindowAC_GasHeaters'
    win_ac = 'WindowAC'


class VRFEquipmentType(str, Enum):
    vrf = 'VRF'


class GasUnitHeaterEquipmentType(str, Enum):
    guh = 'GasHeaters'


class RadiantEquipmentType(str, Enum):
    radiant_chill_gb = 'Radiant_Chiller_Boiler'
    radiant_chill_ashp = 'Radiant_Chiller_ASHP'
    radiant_chill_dhw = 'Radiant_Chiller_DHW'
    radiant_ac_chill_gb = 'Radiant_ACChiller_Boiler'
    radiant_ac_chill_ashp = 'Radiant_ACChiller_ASHP'
    radiant_ac_chill_dhw = 'Radiant_ACChiller_DHW'
    radiant_dcw_gb = 'Radiant_DCW_Boiler'
    radiant_dcw_ashp = 'Radiant_DCW_ASHP'
    radiant_dcw_dhw = 'Radiant_DCW_DHW'


class FCU(_HeatCoolBase):
    """Fan Coil Unit (FCU) heating/cooling system (with no ventilation).

    Each room/zone receives its own Fan Coil Unit (FCU), which meets the heating
    and cooling loads of the space. The cooling coil in the FCU is always chilled
    water cooling coil, which is connected to a chilled water loop operating
    at 6.7C (44F). The heating coil is a hot water coil except when when electric
    baseboards or gas heaters are specified. Hot water temperature is 82C (180F) for
    boiler/district heating and 49C (120F) when ASHP is used.
    """

    type: constr(regex='^FCU$') = 'FCU'

    equipment_type: FCUEquipmentType = Field(
        FCUEquipmentType.fcu_chill_gb,
        description='Text for the specific type of system equipment from the '
        'FCUEquipmentType enumeration.'
    )


class Baseboard(_HeatCoolBase):
    """Baseboard heating system.

    Baseboard systems are intended for spaces only requiring heating and
    no ventilation or cooling. Each room/zone will get its own baseboard
    heating unit that satisfies the heating load.
    """

    type: constr(regex='^Baseboard$') = 'Baseboard'

    equipment_type: BaseboardEquipmentType = Field(
        BaseboardEquipmentType.e_base,
        description='Text for the specific type of system equipment from the '
        'BaseboardEquipmentType enumeration.'
    )


class EvaporativeCooler(_HeatCoolBase):
    """Direct evaporative cooling systems (with optional heating).

    Each room/zone will receive its own air loop sized to meet the sensible load,
    which contains an evaporative cooler that directly adds humidity to the room
    air to cool it. The loop contains an outdoor air mixer, which is used whenever
    the outdoor air has a lower wet bulb temperature than the return air from
    the room. In the event that the combination of outdoor and room return air
    air is too humid, a backup single-speed direct expansion (DX) cooling coil
    will be used. Heating loads can be met with various options, including
    several types of baseboards, a furnace, or gas unit heaters.
    """

    type: constr(regex='^EvaporativeCooler$') = 'EvaporativeCooler'

    equipment_type: EvaporativeCoolerEquipmentType = Field(
        EvaporativeCoolerEquipmentType.evap_e_base,
        description='Text for the specific type of system equipment from the '
        'EvaporativeCoolerEquipmentType enumeration.'
    )


class WSHP(_HeatCoolBase):
    """Water Source Heat Pump (WSHP) heating/cooling system (with no ventilation).

    Each room/zone receives its own Water Source Heat Pump (WSHP), which meets
    the heating and cooling loads of the space. All WSHPs are connected to the
    same water condenser loop, which has its temperature maintained by the
    equipment_type (eg. Boiler with Cooling Tower).
    """

    type: constr(regex='^WSHP$') = 'WSHP'

    equipment_type: WSHPEquipmentType = Field(
        WSHPEquipmentType.wshp_fc_gb,
        description='Text for the specific type of system equipment from the '
        'WSHPEquipmentType enumeration.'
    )


class Residential(_HeatCoolBase):
    """Residential Air Conditioning, Heat Pump or Furnace system.

    Residential HVAC systems are intended primarily for single-family homes and
    include a wide variety of options. In all cases, each room/zone will receive
    its own air loop WITHOUT an outdoor air inlet (air is simply being recirculated
    through the loop). Residential air conditioning (AC) systems are modeled
    using a unitary system with a single-speed direct expansion (DX) cooling
    coil in the loop. Residential heat pump (HP) systems use a single-speed DX
    heating coil in the unitary system and the residential furnace option uses
    a gas coil in the unitary system. In all cases, the properties of these coils
    are set to reflect a typical residential system.
    """

    type: constr(regex='^Residential$') = 'Residential'

    equipment_type: ResidentialEquipmentType = Field(
        ResidentialEquipmentType.res_ac_e_base,
        description='Text for the specific type of system equipment from the '
        'ResidentialEquipmentType enumeration.'
    )


class WindowAC(_HeatCoolBase):
    """Window Air Conditioning cooling system (with optional heating).

    Each room/zone will receive its own Packaged Terminal Air Conditioner (PTAC)
    with properties set to reflect a typical window air conditioning (AC) unit.
    No ventilation air is supplied by the unit and the cooling coil within the
    unit is a single-speed direct expansion (DX) cooling coil. Heating loads
    can be met with various options, including several types of baseboards,
    a furnace, or gas unit heaters.
    """

    type: constr(regex='^WindowAC$') = 'WindowAC'

    equipment_type: WindowACEquipmentType = Field(
        WindowACEquipmentType.win_ac_e_base,
        description='Text for the specific type of system equipment from the '
        'WindowACEquipmentType enumeration.'
    )


class VRF(_HeatCoolBase):
    """Variable Refrigerant Flow (VRF) heating/cooling system (with no ventilation).

    Each room/zone receives its own Variable Refrigerant Flow (VRF) terminal,
    which meets the heating and cooling loads of the space. All room/zone terminals
    are connected to the same outdoor unit, meaning that either all rooms must be
    in cooling or heating mode together.
    """

    type: constr(regex='^VRF$') = 'VRF'

    equipment_type: VRFEquipmentType = Field(
        VRFEquipmentType.vrf,
        description='Text for the specific type of system equipment from the '
        'VRFEquipmentType enumeration.'
    )


class GasUnitHeater(_HeatCoolBase):
    """Gas unit heating system.

    Gas unit systems are intended for spaces only requiring heating and no
    ventilation or cooling. Each room/zone will get its own gaa heating unit
    that satisfies the heating load.
    """

    type: constr(regex='^GasUnitHeater$') = 'GasUnitHeater'

    equipment_type: GasUnitHeaterEquipmentType = Field(
        GasUnitHeaterEquipmentType.guh,
        description='Text for the specific type of system equipment from the '
        'GasUnitHeaterEquipmentType enumeration.'
    )


class Radiant(_HeatCoolBase):
    """Low temperature radiant HVAC system.

    This HVAC template will change the floor and/or ceiling constructions
    of the Rooms that it is applied to, replacing them with a construction that
    aligns with the radiant_type property (eg. CeilingMetalPanel).

    The heating and cooling needs of the space are met with the radiant constructions,
    which use chilled water at 12.8C (55F) and a hot water temperature somewhere
    between 32.2C (90F) and 49C (120F) (warmer temperatures are used in colder
    climate zones).

    Note that radiant systems are particularly limited in cooling capacity and
    using them may result in many unmet hours. To reduce unmet hours, one can
    remove carpets, reduce internal loads, reduce solar and envelope gains during
    peak times, add thermal mass, and use an expanded comfort range.
    """

    type: constr(regex='^Radiant$') = 'Radiant'

    equipment_type: RadiantEquipmentType = Field(
        RadiantEquipmentType.radiant_chill_gb,
        description='Text for the specific type of system equipment from the '
        'RadiantEquipmentType enumeration.'
    )

    radiant_face_type: RadiantFaceTypes = Field(
        RadiantFaceTypes.floor,
        description='Text to indicate which faces are thermally active by default. '
        'Note that this property has no effect when the rooms to which the HVAC '
        'system is assigned have constructions with internal source materials. '
        'In this case, those constructions will dictate the thermally active '
        'surfaces.'
    )

    minimum_operation_time: float = Field(
        1.0,
        gt=0,
        description='A number for the minimum number of hours of operation '
        'for the radiant system before it shuts off.'
    )

    switch_over_time: float = Field(
        24.0,
        gt=0,
        description='A number for the minimum number of hours for when the system '
        'can switch between heating and cooling.'
    )



"""Changes associated with version Honeybee schema version 1.43.2."""
UPDATED_CONSTRUCTS = ('OpaqueConstructionAbridged', 'WindowConstructionAbridged')


def version_1_43_2(model_dict):
    """Implement changes in a Model dict to make it compatible with version 1.43.2."""
    if 'energy' in model_dict['properties']:
        if 'constructions' in model_dict['properties']['energy']:
            for construct in model_dict['properties']['energy']['constructions']:
                if construct['type'] in UPDATED_CONSTRUCTS:
                    construct['materials'] = construct.pop('layers')
    return model_dict



"""Global modifier-set for Model."""
import pathlib
import json

from typing import List, Union
from pydantic import constr, Field

from honeybee_standards import radiance_default

from .._base import NoExtraBaseModel
from .modifierset import WallModifierSetAbridged, FloorModifierSetAbridged, \
    RoofCeilingModifierSetAbridged, ApertureModifierSetAbridged, \
    DoorModifierSetAbridged, ShadeModifierSetAbridged
from .modifier import Plastic, Glass, Trans


# import modifierset default values from honeybee standards
_DEFAULTS = json.loads(pathlib.Path(radiance_default).read_text())
_MOD_SET = [
    ms for ms in _DEFAULTS['modifier_sets']
    if ms['identifier'] == 'Generic_Interior_Visible_Modifier_Set'][0]
_MODIFIER_NAMES = [
    'generic_wall_0.50', 'generic_floor_0.20', 'generic_ceiling_0.80',
    'generic_interior_window_vis_0.88', 'generic_exterior_window_vis_0.64',
    'generic_opaque_door_0.50', 'generic_interior_shade_0.50',
    'generic_exterior_shade_0.35', 'generic_context_0.20', 'air_boundary'
]
_MODIFIERS = [
    Plastic.parse_obj(m) if m['type'] == 'Plastic'
    else Glass.parse_obj(m) if m['type'] == 'Glass'
    else Trans.parse_obj(m)
    for m in _DEFAULTS['modifiers'] if m['identifier'] in _MODIFIER_NAMES
]


class GlobalModifierSet(NoExtraBaseModel):

    type: constr(regex='^GlobalModifierSet$') = 'GlobalModifierSet'

    modifiers: List[Union[Plastic, Glass, Trans]] = Field(
        default=_MODIFIERS,
        description='Global Honeybee Radiance modifiers.',
        readOnly=True
    )

    wall_set: WallModifierSetAbridged = Field(
        default=WallModifierSetAbridged.parse_obj(_MOD_SET['wall_set']),
        description='Global Honeybee WallModifierSet.',
        readOnly=True
    )

    floor_set: FloorModifierSetAbridged = Field(
        default=FloorModifierSetAbridged.parse_obj(_MOD_SET['floor_set']),
        description='Global Honeybee FloorModifierSet.',
        readOnly=True
    )

    roof_ceiling_set: RoofCeilingModifierSetAbridged = Field(
        default=RoofCeilingModifierSetAbridged.parse_obj(_MOD_SET['roof_ceiling_set']),
        description='Global Honeybee RoofCeilingModifierSet.',
        readOnly=True
    )

    aperture_set: ApertureModifierSetAbridged = Field(
        default=ApertureModifierSetAbridged.parse_obj(_MOD_SET['aperture_set']),
        description='Global Honeybee ApertureModifierSet.',
        readOnly=True
    )

    door_set: DoorModifierSetAbridged = Field(
        default=DoorModifierSetAbridged.parse_obj(_MOD_SET['door_set']),
        description='Global Honeybee DoorModifierSet.',
        readOnly=True
    )

    shade_set: ShadeModifierSetAbridged = Field(
        default=ShadeModifierSetAbridged.parse_obj(_MOD_SET['shade_set']),
        description='Global Honeybee ShadeModifierSet.',
        readOnly=True
    )

    air_boundary_modifier: str = Field(
        default=_MOD_SET['air_boundary_modifier'],
        description='Global Honeybee Modifier for AirBoundary Faces.',
        readOnly=True
    )

    context_modifier: str = Field(
        default='generic_context_0.20',
        description='Global Honeybee Modifier for context Shades.',
        readOnly=True
    )
